local addonName = "gnOOM"
local chatShort = "go"
local chatLong = addonName
local addonDbName = addonName .. 'DB'
local addon = LibStub('AceAddon-3.0'):NewAddon(addonName, 'AceConsole-3.0', 'AceEvent-3.0', 'LibDebugLog-1.0', 'LibBars-1.0')
--local ldb   = LibStub:GetLibrary('LibDataBroker-1.1')
local L = LibStub('AceLocale-3.0'):GetLocale(addonName)
--local bars = LibStub('LibBars-1.0')
local _G = _G
_G[addonName] = addon

local options = {
	name = addonName,
	handler = addon,
	type = 'group',
	args = {
 		enable = {
 			name = "Enable",
 			desc = "Enable addon.",
 			type = 'toggle',
 			get = 'getEnable',
 			set = 'setEnable',
 		},
		layout = {
			name = "Layout",
			desc = "Layout options.",
			type = 'group',
			args = {
				anchor = {
					name = "Show anchor",
					desc = "Show anchor. This makes the bars movable.",
					type = 'toggle',
					get = 'getUnlock',
					set = 'setUnlock',
				},
				oOMBar = {
					name = "Show OOM Bar",
					desc = "Show OOM Bar.",
					type = 'toggle',
					get = 'getOOMBar',
					set = 'setOOMBar',
				},
				maxTime = {
					name = "maxTime",
					desc = "maximum time displayed",
					type = 'range',
					min = 1,
					max = 600,
					bigStep = 1,
					get = 'getMaxTime',
					set = 'setMaxTime',
				},
				x = {
					name = "X Position",
					desc = "Horizonzal position of the bars.",
					type = 'range',
					min = 0,
					max = 1000,
					step = 1,
					get = 'getX',
					set = 'setX',
				},
				y = {
					name = "Y Position",
					desc = "Vertical position of the bars.",
					type = 'range',
					min = 0,
					max = 1000,
					step = 1,
					get = 'getY',
					set = 'setY',
				},
				width = {
					name = "Width",
					desc = "Width of the bars.",
					type = 'range',
					min = 10,
					max = 1000,
					step = 1,
					get = 'getWidth',
					set = 'setWidth',
				},
				height = {
					name = "Height",
					desc = "Height of one bar.",
					type = 'range',
					min = 5,
					max = 50,
					step = 1,
					get = 'getHeight',
					set = 'setHeight',
				},
				fontSize = {
					name = "Font size",
					desc = "Font size of the text.",
					type = 'range',
					min = 6,
					max = 48,
					step = 1,
					get = 'getFontSize',
					set = 'setFontSize',
				},
			}, -- layout.args
		}, -- layout
		minPercent = {
			name = "Minimal Percentage",
			desc = "Percent of mana considered OOM.",
			type = 'range',
			min = 0,
			max = 1,
			isPercent = true,
			get = 'getMinPercent',
			set = 'setMinPercent',
		},
		calc = {
			name = "Calc",
			desc = "Calculation options.",
			type = 'group',
			args = {
				algorithm = {
					name = "Algorithm",
					desc = "Choose the algorithm",
					type = 'select',
					values = { 
						exp = "Exponential averaging.",
						avg = "Arithmetic average.",
						cast = "Count the last n casts.",
					},
					get = 'getAlgorithm',
					set = 'setAlgorithm',
				},
				theta = {
					name = "theta",
					desc = "Exponential average coefficient, 1 is fanatic, 0 is ultra conservative.",
					type = 'range',
					min = 0,
					max = 1,
					bigStep = 1,
					isPercent = true,
					get = 'getTheta',
					set = 'setTheta',
				},
				time = {
					name = "time",
					desc = "Window duration for the arithmetic average.",
					type = 'range',
					min = 1,
					max = 60,
					bigStep = 1,
					get = 'getTimeWindow',
					set = 'setTimeWindow',
				},
				numCasts = {
					name = "Number of Casts",
					desc = "Number of Casts for the cast algorithm.",
					type = 'range',
					min = 1,
					max = 10,
					bigStep = 1,
					get = 'getNumCasts',
					set = 'setNumCasts',
				},
				regPercentTreshold = {
					name = "regPercentTreshold",
					desc = "Mana gain over this percentage is ignored.",
					type = 'range',
					min = 0,
					max = 1,
					bigStep = 1,
					isPercent = true,
					get = 'getTreshold',
					set = 'setTreshold',
				},
			}, -- calc.args
		}, -- calc
		
	},
}

--[[
-- Getter Setter block.
--]]
function addon:getEnable(info)
	return self.db.profile.enabled
end

function addon:setEnable(info,value)
	self.db.profile.enabled = value
	if value then
		-- enable
		self:CreateBars()
	else
		-- disable
		if self.oOMBar then
			self.oOMBar:Hide()
		end
	end
end

function addon:getOOMBar(info)
	return self.db.profile.showBar
end

function addon:setOOMBar(info,value)
	self.db.profile.showBar = value
	if value then
		-- show
		self:CreateBars()
	else
		-- hide
-- 		if self.oOMBar then
-- 			self.oOMBar:Hide()
-- 		end
	end
end

function addon:getUnlock(info)
	return self.db.profile.unlock
end

function addon:setUnlock(info,value)
	self.db.profile.unlock = value
	if value and self.bars then
		-- show
		self.bars:ShowAnchor()
	elseif self.bars then
		-- hide
		self.bars:HideAnchor()
	end
end

function addon:getMinPercent()
	return self.db.profile.minPercent
end

function addon:setMinPercent(info,value)
	self.db.profile.minPercent = value
end

function addon:getTheta()
	return self.db.profile.theta
end

function addon:setTheta(info,value)
	self.db.profile.theta = value
end

function addon:getTimeWindow()
	return self.db.profile.timeWindow
end

function addon:setTimeWindow(info,value)
	self.db.profile.timeWindow = value
end

function addon:getNumCasts()
	return self.db.profile.numCasts
end

function addon:setNumCasts(info,value)
	self.db.profile.numCasts = value
end


function addon:getMaxTime()
	return self.db.profile.maxTime
end

function addon:getAlgorithm() 
	return self.db.profile.algorithm
end

function addon:setAlgorithm(info,value)
	if value == 'exp' or value == 'avg' or value == 'cast' then
		self.db.profile.algorithm = value
		self:ResetCalc()
	end
end

function addon:setMaxTime(info,value)
	self.db.profile.maxTime = value
end

function addon:getTreshold()
	return self.db.profile.regPercentTreshold
end

function addon:setTreshold(info,value)
	self.db.profile.regPercentTreshold = value
end

function addon:getX()
	return self.db.profile.xPos
end

function addon:setX(info,value)
	self.db.profile.xPos = value
	if self.bars then
		self.bars:SetPoint('TOPLEFT', UIParent, 'BOTTOMLEFT', self.db.profile.xPos, self.db.profile.yPos)
	end
end

function addon:getY()
	return self.db.profile.yPos
end

function addon:setY(info,value)
	self.db.profile.yPos = value
	if self.bars then
		self.bars:SetPoint('TOPLEFT', UIParent, 'BOTTOMLEFT', self.db.profile.xPos, self.db.profile.yPos)
	end
end

function addon:getWidth()
	return self.db.profile.width
end

function addon:setWidth(info,value)
	self.db.profile.width = value
	if self.bars then
		self.bars:SetWidth(value)
	end
end

function addon:getHeight()
	return self.db.profile.height
end

function addon:setHeight(info,value)
	self.db.profile.height= value
	if self.bars then
		self.bars:SetHeight(value)
	end
end

function addon:getFontSize()
	return self.db.profile.fontSize
end

function addon:setFontSize(info,value)
	self.db.profile.fontSize = value
	if self.bars then
		self.bars:SetFont(nil,value,nil)
	end
-- 	if self.bar then
-- 		self.bar:SetFont(nil,value,nil)
-- 	end
end


local defaults = {
	profile = {
		unlock = false,
		enabled = true,
		minPercent = 0.02,
		theta = 0.1,
		regPercentTreshold = 0.1,
		showBar = true,
		maxTime = 120,
		algorithm = 'avg',
		timeWindow = 20,
		numCasts = 3,
		width = 150,
		height = 12,
		fontSize = 8,
		xPos = 100,
		yPos = 100,
	},
}

--[[
-- Handles chat commands.
--]]
function addon:ChatCommand(input)
    if not input or input:trim() == '' then
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
    else
        LibStub('AceConfigCmd-3.0').HandleCommand(addonName, chatShort, addonName, input)
    end
end

--[[
--	Called when the addon is loaded
--]]
function addon:OnInitialize()
	self.db = LibStub('AceDB-3.0'):New(addonDbName, defaults)
	LibStub('AceConfig-3.0'):RegisterOptionsTable(addonName, options)
	self.optionsFrame = LibStub('AceConfigDialog-3.0'):AddToBlizOptions(addonName, addonName)
	self:RegisterChatCommand(chatShort, 'ChatCommand')
	self:RegisterChatCommand(chatLong, 'ChatCommand')
	do
		local player, _ = UnitName('player')
		if player == 'Chiloe' then
			self:ToggleDebugLog(true)
		end
	end
end

--[[
--	Called when the addon is enabled.
--]]
function addon:OnEnable()
	self:RegisterEvent('UNIT_MANA')
	self:RegisterEvent('PLAYER_REGEN_DISABLED')
	self:RegisterEvent('PLAYER_REGEN_ENABLED')
	self:RegisterEvent('UNIT_SPELLCAST_SUCCEEDED')
	self:CreateBars()
end

function addon:CreateBars()
	if self.db.profile.enabled and not self.bars then
		self.bars = self:NewBarGroup(addonName, nil, self.db.profile.width, self.db.profile.height)
		self.bars:SetFont(nil, self.db.profile.fontSize)
-- 		self.bars:SetColorAt(1.00, 1, 0, 0, 1)
-- 		self.bars:SetColorAt(0.66, 1, 1, 0, 1)
-- 		self.bars:SetColorAt(0.33, 0, 1, 1, 1)
-- 		self.bars:SetColorAt(0.00, 0, 0, 1, 1)
		self.bars:SetPoint('TOPLEFT', UIParent, 'BOTTOMLEFT', self.db.profile.xPos, self.db.profile.yPos)
--		self.bars:SetScale(0.9)
		self.bars.RegisterCallback(self, "AnchorMoved")
		--self.oOMBar = self.bars:NewTimerBar("TTOOM",nil,self.db.profile.maxTime,self.db.profile.maxTime,true)		
		--self.oOMBar:Hide()
		self.bars:HideIcon()
	end
	if self.bars then
		if self.db.profile.unlock then
			self.bars:ShowAnchor()
		else
			self.bars:HideAnchor()
		end
	end
end

function addon:OnDisable()

end

function addon:AnchorMoved(event,id,x,y)
	self.db.profile.xPos = x
	self.db.profile.yPos = y
end

--[[
--  Updates the displayed bar (if enabled)
--
--]]
function addon:UpdateTTOOM( )
	local barName = addonName .. "TTOOM"
	if not self.db.profile.showBar then return end
	--self:Debug("Update")
	local value, maxValue, text, timer
	if self.tTOOM then		
		value = self.tTOOM
		maxValue = self.db.profile.maxTime
		text = nil --date("%M:%S",self.tTOOM)
		timer = true
	else 
 		-- no time => never oom
		value = self.db.profile.maxTime
		maxValue =self.db.profile.maxTime
		text = L["Never"]
		timer = false
	end
	if (not self.oOMBar) or (not self.bars:GetBar(barName)) then
		self.oOMBar = self.bars:NewTimerBar(barName,text,value,maxValue)
	end	
	if timer then
		self.oOMBar:SetTimer(value,maxValue)
		self.oOMBar:ShowTimerLabel()
	else		
		self.oOMBar:StopTimer()
		self.oOMBar:HideTimerLabel()
		self.oOMBar:SetValue(value,maxValue)
	end
	self.oOMBar:SetText(text)
	self.oOMBar:Show()
end

function addon:UNIT_MANA(event, unit)
	if not self.db.profile.enabled then return end
	-- Check for combat and player
	if unit ~= 'player' or not self.combat then return end

	-- Get Current Time (in seconds)
	local time = GetTime()
	local mana = UnitPower('player')
	local maxMana = UnitPowerMax('player')
	local regTreshold = self.db.profile.regPercentTreshold * maxMana
	local minMana = self.db.profile.minPercent * maxMana
		
	if self.db.profile.algorithm == 'exp' then
		
		if (not self.lastTime) or (time-self.lastTime > 10) then
			self:ResetCalc()
			self.lastTime = time;
			self.lastMana = mana
			self.mps = nil
			return
		end
		
		-- Estimate current mps
		local deltaM = mana - self.lastMana
		
		local deltaT = time - self.lastTime
		local mps = deltaM / deltaT
		
		-- Ignore gains larger than treshold
		if (deltaM > 0) and (mps > regTreshold) then
			-- Ignore large gains, so no update of TTOOM
			--self:Debug("Over Treshold.")
		else		
			if self.mps then
				local timedTheta
				if deltaT < 10 then
					timedTheta = math.pow( 1+self.db.profile.theta, deltaT)-1
				else
					timedTheta = math.pow( 1+self.db.profile.theta, 10)-1
				end
				self.mps = timedTheta * mps + (1-timedTheta) * self.mps
				--self:Debug( timedTheta )
			else
				self.mps = mps			
			end
		end
		-- Update last values
		self.lastTime = time
		self.lastMana = mana

	elseif self.db.profile.algorithm == 'avg' or self.db.profile.algorithm == 'cast' then
		if not self.avgReadings then	
			self.avgReadings = {}
		else		
			-- Find closest entry
			local minDist, bestTime, bestMana
			for rTime, rMana in pairs(self.avgReadings) do
				-- Difference beetwen now and then
				local deltaT = time - rTime
				local curDist = math.abs(deltaT - self.db.profile.timeWindow)
				if not minDist or curDist < minDist then
					minDist, bestTime, bestMana = curDist, rTime, rMana
				end
			end
			if bestTime and bestMana then
				local deltaT = time - bestTime
				local deltaM = mana - bestMana
				self.avgMps = deltaM / deltaT
				if self.db.profile.algorithm == 'avg' then
					self.mps = self.avgMps
				end
				--self:Debug( "deltaM", deltaM, "deltaT", deltaT, "mps", self.mps )
			else
				-- no content :(
			end
		end -- if avgReadings
		-- Remove old data
		for rTime, rMana in pairs(self.avgReadings) do
			local deltaT = time - rTime
			-- Allow 1 second more than timeWindow
			if deltaT > self.db.profile.timeWindow+1 then
				self.avgReadings[rTime] = nil
			end
		end
		-- Add new entry
		self.avgReadings[time] = mana
	end
	if self.db.profile.algorithm == 'cast' then
		-- For the cast module we will do the work in another event
		if self.castReadings and time - self.castReadings.lastTime > self.db.profile.timeWindow/2 then
			self.mps = self.avgMps
			--self:Debug("Using avg data for cast")
		else
			return
		end
	end 
	if self.mps then
		if self.mps < 0 then
			self.tTOOM = (mana-minMana) / -self.mps
			self.oOMTime = self.tTOOM + time
		else
			self.tTOOM = nil
			self.oOMTime = nil
		end
		--if mps < 0 or self.mps > 0 then
			self:UpdateTTOOM()
		--end	
	end
end

function addon:UNIT_SPELLCAST_SUCCEEDED(event,unit,name,rank)
	if not self.db.profile.enabled then return end
	-- Check for combat and player
	if unit ~= 'player' or not self.combat then return end

	-- Get Current Time (in seconds)
	local time = GetTime()
	local mana = UnitPower('player')
	local maxMana = UnitPowerMax('player')
	local regTreshold = self.db.profile.regPercentTreshold * maxMana
	local minMana = self.db.profile.minPercent * maxMana
		
	if self.db.profile.algorithm == 'cast' then
		if not self.castReadings then	
			self.castReadings = { top = 0, bottom = -1 }
			function self.castReadings:Push(value)
				local bottom = self.bottom + 1
				self[bottom] = value
				self.bottom = bottom
			end
			
			function self.castReadings:IsEmpty() 
				return self.top <= self.bottom
			end
				
			function self.castReadings:GetSize()
				return self.bottom - self.top + 1
			end
			
			function self.castReadings:Pop()
				local top = self.top
				if top > self.bottom then
					-- This does not need to be localized
					error('List is empty!')
				end
				local value = self[top]
				self[top] = nil
				self.top = top+1
				return value
			end
			
			function self.castReadings:Peek()
				local top = self.top
				if top > self.bottom then
					-- This does not need to be localized
					error('List is empty!')
				end
				return self[top]
			end

			function self.castReadings:Reset()
				self.top = 0
				self.bottom = -1
				for k,v in ipairs(self) do
					self[k] = nil
				end
			end
			
			self.castReadings:Push({ time = time, mana = mana})
			self.castReadings.lastTime = time
			--self:Debug("Created")
		else -- if not castReadings
			-- Find closest entry
			local popped,entry
			do
				local listSize = self.castReadings:GetSize()
				if listSize < 1 then
					self.castReadings:Push({ time = time, mana = mana})
					self.castReadings.lastTime = time
					--self:Debug("Too few data")
					return
				end
				if listSize >= self.db.profile.numCasts then
					entry = self.castReadings:Pop()
					popped = true
					--self:Debug("Popped")
				else
					entry = self.castReadings:Peek()
					popped = false
					--self:Debug("Peeked")
				end
			end
						
			local deltaT = time - entry.time
			local deltaM = mana - entry.mana
			self.mps = deltaM / deltaT

			if popped then
				entry.time = time
				entry.mana = mana
				self.castReadings:Push(entry)
				--self:Debug("Recycled")
			else
				self.castReadings:Push({ time = time, mana = mana})
				--self:Debug("New Entry")
			end
			self.castReadings.lastTime = time
		end -- if not castReadings
		-- Add new entry
		
	else
		-- None of these algorithms, ignore
		return
	end -- case
	if self.mps then
		if self.mps < 0 then
			self.tTOOM = (mana-minMana) / -self.mps
			self.oOMTime = self.tTOOM + time
		else
			self.tTOOM = nil
			self.oOMTime = nil
		end
		self:UpdateTTOOM()
	end

end

function addon:PLAYER_REGEN_DISABLED(...)
	if self.db.profile.enabled then
		self.combat = true
	end
end


function addon:PLAYER_REGEN_ENABLED(...)
	if self.db.profile.enabled then
		self.combat = false
		self:ResetCalc()
	end
end

function addon:ResetCalc() 
	self.lastMana = nil
	self.lastTime = nil
	self.mps = nil
	self.tTOOM = nil
	self.oOMTime = nil
	if self.avgReadings then
		wipe(self.avgReadings)
	end
	if self.castReadings then
		self.castReadings:Reset()
	end
	if self.oOMBar then
		self.oOMBar:Hide()
	end
end